import Foundation

extension OutputStream {
    enum WriteResult {
        case completed
        case partial(unwrittenRemainder: Data)
        case failure(error: Error)
    }

    func write(buffer: Data) -> WriteResult {
        let bytesAvailable = buffer.count
        return buffer.withUnsafeBytes {
            guard let rawPointer = $0.baseAddress, $0.count > 0 else {
                // A no-data write completes trivially.
                return .completed
            }

            let bytesWritten = write(rawPointer.assumingMemoryBound(to: UInt8.self), maxLength: bytesAvailable)
            if bytesWritten < 0 {
                return .failure(error: streamError!)
            } else if bytesWritten == bytesAvailable {
                return .completed
            } else {
                return .partial(unwrittenRemainder: buffer.suffix(from: bytesWritten))
            }
        }
    }
}

extension InputStream {
    enum ReadResult {
        case readSome(slice: Slice<UnsafeMutableBufferPointer<UInt8>>)
        case doneReading
        case failedReading(error: Error?)
    }

    func read(buffer: UnsafeMutableBufferPointer<UInt8>) -> ReadResult {
        let spaceAvailable = buffer.count
        precondition(spaceAvailable > 0, "Cannot read into buffer of size \(spaceAvailable)")
        let address = buffer.baseAddress!
        let bytesRead = read(address, maxLength: spaceAvailable)
        if bytesRead < 0 {
            return .failedReading(error: streamError)
        } else if bytesRead == 0 && streamStatus == .atEnd {
            return .doneReading
        } else {
            return .readSome(slice: buffer.prefix(bytesRead))
        }
    }
}
