import Foundation

final class StreamingTransformer: NSObject {
    /// The actual transform working on the raw bytes, producing the data we want to stream, e.g. to a server.
    ///
    /// The buffer passed to the closure needs to be consumed in its entirety, and must not leave its scope. If an
    /// error occurs, it should be returned — processing will stop, and the underlying stream(s) will be closed.
    typealias Transform = (UnsafeBufferPointer<UInt8>) -> Result<Data, Error>

    /// Creates a new transformer that streams the bytes of `file`, using the user-supplied transform.
    class func transforming(file: URL, transform: @escaping Transform) throws -> StreamingTransformer {
        precondition(file.isFileURL, "Can only work on file URLs")
        guard let actualSource = InputStream(url: file) else {
            throw Errors.cannotOpen(file: file, underlyingError: nil)
        }

        actualSource.open()
        if actualSource.streamStatus != .open {
            throw Errors.cannotOpen(file: file, underlyingError: actualSource.streamError)
        }
        let bufferSize = 512
        let pair = try makeOpenedStreamPair(bufferSize: bufferSize)

        return .init(source: actualSource, bufferSize: bufferSize, boundPair: pair, transform: transform)
    }
    
    /// Begins streaming the underlying bytes on the given runloop, while data is read from from the returned stream.
    ///
    /// While streaming, the `transform` parameter passed on creation will be invoked until either:
    /// - the source of this object has been streamed in its entirety
    /// - an error is encountered while reading from the source
    /// - an error is returned from the `transform`
    /// - the returned `InputStream` is closed or encounters an error
    ///
    /// Whichever happens first.
    ///
    /// - Parameters:
    ///   - runloop: The runloop on which the data shall be read from the source, and the `transform` is called.
    /// - Throws: Throws an error if the object has already begun streaming.
    /// - Returns: Returns an opened `InputStream` that the tranformed bytes can be read from.
    func beginStreaming(in runloop: RunLoop) throws -> InputStream {
        switch state {
        case .done:
            throw Errors.sourceAlreadyExhausted
        case .streaming:
            throw Errors.alreadyStreaming
        case .idle:
            state = .streaming(in: runloop, for: .default)
            drain.schedule(in: runloop, forMode: .default)
        }

        return sourceToVend
    }

    private static func makeOpenedStreamPair(bufferSize: Int) throws -> (source: InputStream, drain: OutputStream) {
        var maybeInput: InputStream?
        var maybeOutput: OutputStream?
        Stream.getBoundStreams(withBufferSize: bufferSize, inputStream: &maybeInput, outputStream: &maybeOutput)
        guard let input = maybeInput, let output = maybeOutput else {
            throw Errors.cannotAllocateBuffer(size: bufferSize)
        }
        input.open()
        output.open()
        assert(input.streamStatus == .open, "Could not open reading end of bound stream pair")
        assert(output.streamStatus == .open, "Could not open writing end of bound stream pair")

        return (source: input, drain: output)
    }

    private init(source: InputStream, bufferSize: Int, boundPair: (source: InputStream, drain: OutputStream), transform: @escaping Transform) {
        precondition(bufferSize > 0, "Cannot read buffers with a size of 0")
        self.actualSource = source
        self.readBuffer = .allocate(capacity: bufferSize)
        self.drain = boundPair.drain
        self.sourceToVend = boundPair.source
        self.bufferTransform = transform
        self.state = .idle
        super.init()
        self.drain.delegate = self
    }

    deinit {
        readBuffer.deallocate()
        drain.delegate = nil
        if case .streaming(let runloop, let mode) = state {
            drain.remove(from: runloop, forMode: mode)
        }
        actualSource.close()
        drain.close()
        sourceToVend.close()
    }

    private let bufferTransform: Transform
    private let readBuffer: UnsafeMutableBufferPointer<UInt8>
    private var leftover: Data?
    private let drain: OutputStream
    private let sourceToVend: InputStream
    private let actualSource: InputStream
    private var state: State
    enum State {
        case idle, done
        case streaming(in: RunLoop, for: RunLoop.Mode)
    }

    enum Errors: Error {
        case cannotOpen(file: URL, underlyingError: Error?)
        case cannotAllocateBuffer(size: Int)
        case cannotRead(underlyingError: Error?)
        case sourceAlreadyExhausted, alreadyStreaming
        case consumerClosed(bytesRead: Int64?)
    }
}

extension StreamingTransformer: StreamDelegate {
    func stream(_ aStream: Stream, handle eventCode: Stream.Event) {
        if eventCode.contains(.errorOccurred) {
            return handle(error: aStream.streamError)
        }
        if eventCode.contains(.endEncountered) {
            return handleEnd()
        }
        if eventCode.contains(.hasSpaceAvailable) {
            drainBytes()
        } else {
            print("Encountered event: \(eventCode)")
        }
    }

    private func drainBytes() {
        // Check whether we should still be streaming — events likely happen on separate threads.
        guard case .streaming = state else {
            return
        }

        // First clear out anything we didn’t manage to write the last time around
        if let pending = leftover {
            switch drain.write(buffer: pending) {
            case .failure(error: let error):
                // Write failed, no point in clearing the leftover.
                return handle(error: error)
            case .partial(unwrittenRemainder: let remainder):
                // Remember the remainder so we can write it the next time the drain has space available.
                leftover = remainder
                return
            case .completed:
                // Entirety of the leftover buffer has been written => clear it and proceed!
                leftover = nil
            }
        }
        assert(leftover == nil, "Flushing should have cleared leftovers or stated we need to bail")

        while drain.hasSpaceAvailable {
            // Fetch fresh bytes from the source, and transform them.
            let readResult = actualSource.read(buffer: readBuffer)
            let transformResult: Result<Data, Error>
            switch readResult {
            case .failedReading(error: let error):
                return handle(error: error)
            case .doneReading:
                transformResult = bufferTransform(.init(start: nil, count: 0))
            case .readSome(slice: let slice):
                transformResult = bufferTransform(.init(rebasing: slice))
            }

            // If the transform failed, we bail with an error. If, however, it did succed, we write the data.
            switch transformResult {
            case .failure(let error):
                return handle(error: error)
            case .success(let data):
                // Again, writing may fail, or only write _some_ of the bytes — in both cases we bail.
                switch drain.write(buffer: data) {
                case .failure(error: let error):
                    return handle(error: error)
                case .partial(unwrittenRemainder: let remainder):
                    // Remember the remainder so we can flush it the next time there is some space.
                    leftover = remainder
                    return
                case .completed:
                    // And lastly, if we’re done reading and managed to write everything that’s left, we’re done.
                    if case .doneReading = readResult {
                        return finishProcessing()
                    }
                }
            }
        }
    }

    private func handle(error: Error?) {
        guard case .streaming = state else {
            return
        }

        if let error = error {
            print("Oh noes: \(error)") // You would normally publish the specific error here…
        } else {
            print("Something unknown went wrong") // …and a generic error here.
        }
        finishProcessing(publishCompletion: false)
    }

    private func handleEnd() {
        guard case .streaming = state else {
            return
        }

        if actualSource.streamStatus == .atEnd {
            finishProcessing()
        } else {
            let bytesRead = actualSource.property(forKey: .fileCurrentOffsetKey) as? Int64
            handle(error: Errors.consumerClosed(bytesRead: bytesRead))
        }
    }

    private func finishProcessing(publishCompletion: Bool = true) {
        guard case .streaming(in: let loop, for: let mode) = state else {
            return
        }

        print("finishing...")
        // Avoid getting further callbacks.
        drain.delegate = nil
        drain.remove(from: loop, forMode: mode)

        // Update the state.
        state = .done

        // Close the streams.
        actualSource.close()
        drain.close()

        if publishCompletion {
            // Whatever makes sense to publish a successful completion.
        }
    }
}

extension StreamingTransformer.State: CustomDebugStringConvertible {
    var debugDescription: String {
        switch self {
        case .idle:
            return "idle"
        case .done:
            return "done"
        case .streaming(in: let loop, for: let mode):
            return "streaming in \(loop) (Mode: \(mode)"
        }
    }
}
